@@ -29,14 +29,70 @@ from scipy.sparse import coo_matrix, csr_matrix, lil_matrix
 from scipy.sparse.linalg import eigsh
 from scipy.linalg import eig
 import scipy.sparse as sparse
+from scipy.sparse import issparse
 import copy
 import random
-import multiprocessing as mp
 import pandas as pd
 import sys
 from hicstuff.log import logger
 
 
+def distance_law_from_mat(matrix, indices=None, log_bins=True, base=1.1):
+    """Compute distance law as a function of the genomic coordinate aka P(s).
+    Bin length increases exponentially with distance if log_bins is True. Works
+    on dense and sparse matrices. Less precise than the one from the pairs.
+    Parameters
+    ----------
+    matrix : numpy.array or scipy.sparse.coo_matrix
+        Hi-C contact map of the chromosome on which the distance law is
+        calculated.
+    indices : None or numpy array
+        List of indices on which to compute the distance law. For example
+        compartments or expressed genes.
+    log_bins : bool
+        Whether the distance law should be computed on exponentially larger
+        bins.
+    Returns
+    -------
+    numpy array of floats :
+        The start index of each bin.
+    numpy array of floats :
+        The distance law computed per bin on the diagonal
+    """
+
+    n = min(matrix.shape)
+    included_bins = np.zeros(n, dtype=bool)
+    if indices is None:
+        included_bins[:] = True
+    else:
+        included_bins[indices] = True
+    D = np.array(
+        [
+            np.average(matrix.diagonal(j)[included_bins[: n - j]])
+            for j in range(n)
+        ]
+    )
+    if not log_bins:
+        return np.array(range(len(D))), D
+    else:
+        n_bins = int(np.log(n) / np.log(base) + 1)
+        logbin = np.unique(
+            np.logspace(0, n_bins - 1, num=n_bins, base=base, dtype=np.int)
+        )
+        logbin = np.insert(logbin, 0, 0)
+        logbin[-1] = min(n, logbin[-1])
+        if n < logbin.shape[0]:
+            print("Not enough bins. Increase logarithm base.")
+            return np.array(range(len(D))), D
+        logD = np.array(
+            [
+                np.average(D[logbin[i - 1] : logbin[i]])
+                for i in range(1, len(logbin))
+            ]
+        )
+        return logbin[:-1], logD
+
+
 def despeckle_simple(B, th2=2, threads=1):
     """Single-chromosome despeckling
 
@@ -71,29 +127,15 @@ def despeckle_simple(B, th2=2, threads=1):
     stds = np.zeros(n1)
     # Faster structure for editing values
     A = lil_matrix(A)
-
-    # only global functions can be used with Pool.map
-    global _diagstats
-
-    def _diagstats(u):
-        """Computes median and standard deviation for each diagonal"""
+    for u in range(n1):
         diag = B.diagonal(u)
         medians[u] = np.median(diag)
-        stds[u] = np.median(diag)
-
-    global _speck2med
-
-    def _speck2med(nw):
-        """Sets outlier values back to the median of their diagonal"""
+        stds[u] = np.std(diag)
+    for nw in range(n1):
         diag = A.diagonal(nw)
         diag[diag > medians[nw] + th2 * stds[nw]] = medians[nw]
         A.setdiag(diag, nw)
 
-    pool = mp.Pool(threads, maxtasksperchild=10)
-    pool.map(_diagstats, range(n1))
-    pool.map(_speck2med, range(n1))
-    pool.close()
-
     return csr_matrix(A)
 
 
@@ -256,7 +298,7 @@ def build_pyramid(M, subsampling_factor=3):
 
 def bin_bp_dense(M, positions, bin_len=10000):
     """Perform binning with a fixed genomic length in
-    base pairs. Fragments will be binned such
+    base pairs on a dense matrix. Fragments will be binned such
     that their total length is closest to the specified input.
     If a contig list is specified, binning will be performed
     such that fragments never overlap two contigs. Fragments longer
@@ -280,10 +322,46 @@ def bin_bp_dense(M, positions, bin_len=10000):
     numpy array of ints :
         List of binned fragments positions in basepair
     """
-    positions = np.array(positions)
+    # Just converting to sparse and passing to sparse function
+    S = coo_matrix(M)
+    out_S, out_pos = bin_bp_sparse(S, positions, bin_len=bin_len)
+    out_M = out_S.todense()
+
+    return out_M, out_pos
+
+
+def bin_bp_sparse(M, positions, bin_len=10000):
+    """
+    Perform binning with a fixed genomic length in
+    base pairs on a sparse matrix. Fragments will be binned such
+    that their total length is closest to the specified input.
+    If a contig list is specified, binning will be performed
+    such that fragments never overlap two contigs. Fragments longer
+    than bin size will not be split, which can result in larger bins.
+    The last smaller bin of the chromosome will be merged with the
+    previous one.
+
+    Parameters
+    ----------
+    M : sparse numpy matrix
+        Hi-C contact matrix in sparse format.
+    positions : numpy array of ints
+        Start positions of fragments in the matrix, in base pairs.
+    bin_len : int
+        Desired length of bins, in base pairs
+
+    Returns
+    -------
+    sparse scipy sparse coo_matrix:
+        The binned sparse matrix in COO format.
+    list of ints:
+        The new bin start positions.
+    """
+
+    r = M.tocoo()
     # Get fragments where new chromosome starts (positions reset)
     chromstart = np.where(positions == 0)[0]
-    chromend = np.append(chromstart[1:], M.shape[0])
+    chromend = np.append(chromstart[1:], len(positions))
     chromlen = chromend - chromstart
     # Assign a chromosome to each fragment
     chroms = np.repeat(range(len(chromlen)), chromlen)
@@ -292,28 +370,70 @@ def bin_bp_dense(M, positions, bin_len=10000):
     frags = np.array([chroms, positions], dtype=np.int64).T
     # Keep track of index fragments
     frag_idx = range(frags.shape[0])
-    # Number of bins to create
-    n_bins = np.unique(frags, axis=0).shape[0]
+    # Unique bin coordinates to create
+    unique_bins = np.unique(frags, axis=0)
+    # Check if some bins are missing (happens if a single
+    # fragment should contain multiple bins)
+    bins_jumps =  (unique_bins[1:, 1] - unique_bins[:-1, 1]) - 1
+    # Compute number of missing bins to add (no restriction site in bin)
+    missing_bins = np.where(bins_jumps > 0)[0]
+    n_missing_bins = np.sum(bins_jumps[bins_jumps>0])
+    # Compute correct number of bins to create
+    n_bins = unique_bins.shape[0] + n_missing_bins
     # Initialise output fragment list (post binning)
     out_pos = np.zeros((n_bins, 1))
-    # initialise matrix for matching frags (row) to bins (col)
-    bin_attr = np.zeros((frags.shape[0], n_bins))
-    # Use (chr, bin#) as grouping key (coord) and indices of fragments
+    row = copy.copy(r.row)
+    col = copy.copy(r.col)
+    # unique_bin_No: Number of bins w/ unique restriction fragments
+    # actual_bin_No: Number of bins in total (including missing ones 
+    # sharing the same fragment)
+    unique_bin_No, actual_bin_No = 0, 0
+    # Match empty missing bins added with the original bin sharing
+    # the same fragment
+    added_bins = {}
+    bin_per_frag = {}
+    # Use (chr, bin) as grouping key (coord) and indices of fragments
     # belonging to current bin (bin_frags)
-    bin_No = 0
     for coords, bin_frags in itertools.groupby(
         frag_idx, lambda x: tuple(frags[x, :])
     ):
+
         bin_frags = list(bin_frags)
         first_frag, last_frag = bin_frags[0], bin_frags[-1] + 1
-        out_pos[bin_No] = coords[1] * bin_len
-        # Fill frag-bin attribution matrix matrix line by line
-        bin_attr[first_frag:last_frag, bin_No] = 1
-        bin_No += 1
-    # Perform binning (sum of contacts in each bin) using dot products
-    D = M.dot(bin_attr)
-    out_M = bin_attr.T.dot(D)
-    return out_M, out_pos
+        # Pool row/col number by bin
+        row[np.where((r.row >= first_frag) & (r.row < last_frag))] = actual_bin_No
+        col[np.where((r.col >= first_frag) & (r.col < last_frag))] = actual_bin_No
+        # Get bin position in basepair
+        out_pos[actual_bin_No] = coords[1] * bin_len
+        # Multiple bins to create in same fragment (rare)
+        if unique_bin_No in missing_bins:
+            # Number of basepairs to shift inside fragment for new bins
+            curr_shift = 0
+            # Subsequent bins belong to same frag as this one
+            orig_bin = copy.copy(actual_bin_No)
+            # Shifting bin index (introducing empty bins) for each bin in same frag
+            for _ in range(bins_jumps[unique_bin_No]):
+                curr_shift += bin_len
+                actual_bin_No += 1
+                # Remember bin coords and #bin /frag to fill contacts later
+                added_bins[actual_bin_No] = orig_bin
+                bin_per_frag[orig_bin] = bin_per_frag.get(
+                    unique_bin_No,
+                    0
+                ) + 1
+                out_pos[actual_bin_No] = coords[1] * bin_len + curr_shift
+        unique_bin_No += 1
+        actual_bin_No += 1
+    row[np.where(r.row >= last_frag)] = actual_bin_No
+    col[np.where(r.col >= last_frag)] = actual_bin_No
+    # Sum data of duplicate row/col pairs 
+    # (i.e. combine contacts of all fragments in same bin)
+
+    binned = coo_matrix((r.data, (row, col)), shape=(actual_bin_No, actual_bin_No))
+    binned.sum_duplicates()
+    binned.eliminate_zeros()
+
+    return (binned, out_pos)
 
 
 def bin_exact_bp_dense(M, positions, bin_len=10000):
@@ -360,65 +480,91 @@ def bin_exact_bp_dense(M, positions, bin_len=10000):
     return N
 
 
-def bin_bp_sparse(M, positions, bin_len=10000):
+def mad(M, axis=None):
     """
-    Performs the bp-binning procedure on a sparse matrix.
+    Computes median absolute deviation of matrix bins sums.
 
     Parameters
     ----------
-    M : sparse numpy matrix
-        Hi-C contact matrix in sparse format.
-    positions : numpy array of ints
-        Start positions of fragments in the matrix, in base pairs.
-    bin_len : int
-        Desired length of bins, in base pairs
+    M : scipy sparse coo_matrix
+        Sparse matrix in COO format.
 
+    axis: int
+        Compute MAD on rows if 0, on columns if 1 or on all pixels if None.
+        If axis is None, MAD is computed only on nonzero pixels.
     Returns
     -------
-    sparse numpy matrix:
-        The binned matrix in sparse format.
-    list of ints:
-        The new bin start positions.
+    float:
+        MAD estimator of matrix bin sums
     """
+    # Compute median on nonzero data values
+    # otherwise, median is 0 if sufficiently sparse
+    if axis is None:
+        if issparse(M):
+            r = M.tocoo()
+            dist = r.data
+        else:
+            dist = M
+        
+    else:
+        if axis < 0:
+            axis += 2
+        dist = np.array(M.sum(axis=axis, dtype=np.float)).flatten()
 
+    return np.median(np.absolute(dist - np.median(dist)))
+
+
+def get_good_bins(M, n_mad=2.0, s_min=None, s_max=None, symmetric=False):
+    """
+    Filters out bins with outstanding sums using median and MAD
+    of the log transformed distribution of bin sums. Only filters
+    weak outlier bins unless `symmetric` is set to True.
+
+    Parameters
+    ----------
+    M : scipy sparse coo_matrix
+        Input sparse matrix representing the Hi-C contact map.
+
+    n_mad : float
+        Minimum number of median absolut deviations around median in the
+        bin sums distribution at which bins will be filtered out.
+
+    s_min : float
+        Optional fixed threshold value for bin sum below which bins should
+        be filtered out.
+
+    s_max: float
+        Optional fixed threshold value for bin sum above which bins should
+        be filtered out.
+    symmetric : bool
+        If set to true, filters out outliers on both sides of the distribution.
+        Otherwise, only filters out bins on the left side (weak bins).
+    Returns
+    -------
+    numpy array of bool :
+        A 1D numpy array whose length is the number of bins in the matrix and
+        values indicate if bins values are within the acceptable range (1)
+        or considered outliers (0).
+    """
     r = M.tocoo()
-    # Get fragments where new chromosome starts (positions reset)
-    chromstart = np.where(positions == 0)[0]
-    chromend = np.append(chromstart[1:], len(positions))
-    chromlen = chromend - chromstart
-    # Assign a chromosome to each fragment
-    chroms = np.repeat(range(len(chromlen)), chromlen)
-    # Get binned positions
-    positions = positions // bin_len
-    frags = np.array([chroms, positions], dtype=np.int64).T
-    # Keep track of index fragments
-    frag_idx = range(frags.shape[0])
-    # Number of bins to create
-    n_bins = np.unique(frags, axis=0).shape[0]
-    # Initialise output fragment list (post binning)
-    out_pos = np.zeros((n_bins, 1))
-    row = copy.copy(r.row)
-    col = copy.copy(r.col)
-    bin_No = 0
-    # Use (chr, bin) as grouping key (coord) and indices of fragments
-    # belonging to current bin (bin_frags)
-    for coords, bin_frags in itertools.groupby(
-        frag_idx, lambda x: tuple(frags[x, :])
-    ):
-        bin_frags = list(bin_frags)
-        first_frag, last_frag = bin_frags[0], bin_frags[-1] + 1
-        # Pool row/col number by bin
-        row[np.where((r.row >= first_frag) & (r.row < last_frag))] = bin_No
-        col[np.where((r.col >= first_frag) & (r.col < last_frag))] = bin_No
-        # Get bin position in basepair
-        out_pos[bin_No] = coords[1] * bin_len
-        bin_No += 1
+    with np.errstate(divide="ignore", invalid="ignore"):
+        bins = sum_mat_bins(r)
+        bins[bins==0] = 1
+        norm = np.log10(bins)
+        median = np.median(norm)
+        sigma = 1.4826 * mad(norm)
 
-    # Sum data of duplicate row/col pairs
-    binned = coo_matrix((r.data, (row, col)), shape=(bin_No, bin_No))
-    binned.sum_duplicates()
-    return (binned, out_pos)
+    if s_min is None:
+        s_min = median - n_mad * sigma
+    if s_max is None:
+        s_max = median + n_mad * sigma
 
+    if symmetric:
+        filter_bins = (norm > s_min) * (norm < s_max)
+    else:
+        filter_bins = norm > s_min
+
+    return filter_bins
 
 def trim_dense(M, n_std=3, s_min=None, s_max=None):
     """By default, return a matrix stripped of component
@@ -449,59 +595,10 @@ def trim_dense(M, n_std=3, s_min=None, s_max=None):
         The input matrix, stripped of outlier component vectors.
     """
 
-    M = np.array(M)
-    sparsity = M.sum(axis=1)
-    mean = np.mean(sparsity)
-    std = np.std(sparsity)
-    if s_min is None:
-        s_min = mean - n_std * std
-    if s_max is None:
-        s_max = mean + n_std * std
-    elif s_max == 0:
-        s_max = np.amax(M)
-    f = (sparsity > s_min) * (sparsity < s_max)
-    N = M[f][:, f]
-    return N
-
-
-def get_good_bins(M, n_std=3, s_min=None, s_max=None):
-    """
-    Returns a boolean mask marking the outlier bins of a Hi-C contact map with
-    0 and the good bins with 1. Good bins are defined either by a fixed
-    or a number of standard deviations from the mean of all bins values.
-
-    Parameters
-    ----------
-    M : scipy.sparse.coo_matrix
-        Sparse Hi-C contact map
-    n_std : int
-        Minimum number of standard deviation by which a the sum of
-        contacts in a component vector must deviate from the mean
-        to be trimmed.
-    s_min : float
-        Fixed minimum value below which the component vectors will
-        be trimmed.
-    s_max : float
-        Fixed maximum value above which the component vectors will
-        be trimmed.
-
-    Returns
-    -------
-    numpy.array of bool :
-        A 1D numpy array whose length is the number of bins in the matrix and
-        values indicate if bins mean values are within the acceptable range (1)
-        or considered outliers (0).
-    """
-    r = M.tocoo()
-    sparsity = np.array(r.sum(axis=1, dtype=np.float)).flatten()
-    mean = np.mean(sparsity)
-    std = np.std(sparsity)
-    if s_min is None:
-        s_min = mean - n_std * std
-    if s_max is None:
-        s_max = mean + n_std * std
-    f = (sparsity > s_min) * (sparsity < s_max)
-    return f
+    S = coo_matrix(M)
+    S_out = trim_sparse(S, n_std=n_std, s_min=s_min, s_max=s_max)
+    M_out = S_out.todense()
+    return M_out
 
 
 def trim_sparse(M, n_std=3, s_min=None, s_max=None):
@@ -530,15 +627,16 @@ def trim_sparse(M, n_std=3, s_min=None, s_max=None):
     r = M.tocoo()
     f = get_good_bins(M, n_std, s_min, s_max)
     miss_bins = np.cumsum(1 - f)
-    # Mapping pre- and post- trimming indices of bins
+    # Mapping pre- and post- trimming indices of bins, post = -1 means delete
     # Note: There is probably a more efficient way than a dictionary for that
     miss_map = {old: old - offset for old, offset in enumerate(miss_bins)}
+    # Indices of cells that will be kept
     indices = np.where(f[r.row] & f[r.col])
     # Remove sparse rows and shift indices accordingly
     rows = [miss_map[i] for i in r.row[indices]]
     cols = [miss_map[j] for j in r.col[indices]]
     data = r.data[indices]
-    size = max(max(rows), max(cols)) + 1
+    size = max(max(rows, default=-1), max(cols, default=-1)) + 1
     N = coo_matrix((data, (rows, cols)), shape=(size, size))
     return N
 
@@ -547,6 +645,7 @@ def normalize_dense(M, norm="SCN", order=1, iterations=3):
     """Apply one of the many normalization types to input dense
     matrix. Will also apply any callable norms such as a user-made
     or a lambda function.
+    NOTE: Legacy function for dense maps
 
     Parameters
     ----------
@@ -606,10 +705,6 @@ def normalize_dense(M, norm="SCN", order=1, iterations=3):
             s_norm = np.tensordot(s_norm_x, s_norm_y, axes=0)
             s[s_norm != 0] = 1.0 * s[s_norm != 0] / s_norm[s_norm != 0]
 
-    elif norm == "global":
-        s_norm = np.linalg.norm(s, ord=floatorder)
-        s /= 1.0 * s_norm
-
     elif callable(norm):
         s = norm(M)
 
@@ -619,22 +714,23 @@ def normalize_dense(M, norm="SCN", order=1, iterations=3):
     return (s + s.T) / 2
 
 
-def normalize_sparse(M, norm="SCN", order=1, iterations=3):
+def normalize_sparse(M, norm="ICE", iterations=40, n_mad=3.0):
     """Applies a normalization type to a sparse matrix.
 
     Parameters
     ----------
     M : scipy.sparse.csr_matrix of floats
     norm : str or callable
-        Normalization procedure to use. Can be one of "SCN",
-        "mirnylib", "frag" or "global". Can also be a user-
-        defined function.
-    order : int
-        Defines the type of vector norm to use. See numpy.linalg.norm
-        for details.
+        Normalization procedure to use. Can be one of "SCN" or
+        "ICE". Can also be a user-defined function.
     iterations : int
         Iterations parameter when using an iterative normalization
         procedure.
+    n_mad : float
+        Maximum number of median absolute deviations of bin sums to allow for 
+        including bins in the normalization procedure. Bins more than `n_mad`
+        mads below the median are excluded. Bins excluded from normalisation
+        are set to 0.
 
     Returns
     -------
@@ -643,31 +739,36 @@ def normalize_sparse(M, norm="SCN", order=1, iterations=3):
     """
     # Making full symmetric matrix if not symmetric already (e.g. upper triangle)
     r = csr_matrix(M)
-    if (abs(r - r.T) > 1e-10).nnz != 0:
-        r += r.T
-        r.setdiag(r.diagonal() / 2)
-        r.eliminate_zeros()
-    r = r.tocoo()
-    if norm == "SCN":
-        for _ in range(1, iterations):
-            row_sums = np.array(r.sum(axis=1)).flatten()
-            col_sums = np.array(r.sum(axis=0)).flatten()
-            row_indices, col_indices = r.nonzero()
-            r.data /= row_sums[row_indices] * col_sums[col_indices]
-        row_sums = np.array(r.sum(axis=1)).flatten()
+    r = r.astype(np.float64)
+    good_bins = get_good_bins(M, n_mad=n_mad)
+    # Set values in non detectable bins to 0
+    r[~good_bins, :] = 0
+    r[:, ~good_bins] = 0
+    r = coo_matrix(r)
+    r.eliminate_zeros()
+    if norm == "ICE":
+        # Row and col indices of each nonzero value in matrix
+        row_indices, col_indices = r.nonzero()
+        for _ in range(iterations):
+            # Symmetric matrix: rows and cols have identical sums
+            bin_sums = sum_mat_bins(r)
+            # Normalize bin sums by the median sum of detectable bins for stability
+            bin_sums /= np.median(bin_sums[good_bins])
+            # Divide each nonzero value by the product of the sums of
+            # their respective rows and columns.
+            r.data /= bin_sums[row_indices] * bin_sums[col_indices]
+        bin_sums = sum_mat_bins(r)
         # Scale to 1
-        r.data = r.data * (1 / np.mean(row_sums))
-
-    elif norm == "global":
-        try:
-            from scipy.sparse import linalg
-
-            r = linalg.norm(M, ord=order)
-        except (ImportError, AttributeError) as e:
-            logger.error(
-                "I can't import linalg tools for sparse matrices."
-                "Please upgrade your scipy version to 0.16.0."
-            )
+        r.data = r.data * (1 / np.median(bin_sums[good_bins]))
+    elif norm == "SCN":
+        # Similar to ICE, but division is done sequentially by row and then column
+        # sums instead of using product.
+        row_indices, col_indices = r.nonzero()
+        for i in range(iterations):
+            bin_sums = sum_mat_bins(r)
+            r.data /= bin_sums[row_indices]
+            bin_sums = sum_mat_bins(r)
+            r.data /= bin_sums[col_indices]
 
     elif callable(norm):
         r = norm(M)
@@ -678,6 +779,27 @@ def normalize_sparse(M, norm="SCN", order=1, iterations=3):
     return r
 
 
+def sum_mat_bins(mat):
+    """
+    Compute the sum of matrices bins (i.e. rows or columns) using
+    only the upper triangle, assuming symmetrical matrices.
+
+    Parameters
+    ----------
+    mat : scipy.sparse.csr_matrix
+        Contact map in sparse format, either in upper triangle or
+        full matrix.
+    
+    Returns
+    -------
+    numpy.array :
+        1D array of bin sums.
+    """
+    # Equivalaent to row or col sum on a full matrix
+    # Note: mat.sum returns a 'matrix' object. A1 extracts the 1D flat array
+    # from the matrix
+    return mat.sum(axis=0).A1 + mat.sum(axis=1).A1 - mat.diagonal(0)
+
 def GC_partial(portion: str):
     """Manually compute GC content percentage in a DNA string, taking
     ambiguous values into account (according to standard IUPAC notation).
@@ -1100,6 +1222,28 @@ def to_pdb(
             f.write(line)
 
 
+def matrix_to_pdb(
+    matrix,
+    filename,
+    contigs=None,
+    annotations=None,
+    indices=None,
+    special_bins=None,
+    alpha=1,
+):
+    """Convert a matrix to a PDB file, shortcutting the intermediary generated
+    structure.
+    """
+    to_pdb(
+        to_structure(matrix, alpha=alpha),
+        filename=filename,
+        contigs=contigs,
+        annotations=annotations,
+        indices=indices,
+        special_bins=special_bins,
+    )
+
+
 def to_distance(matrix, alpha=1):
     """Compute distance matrix from contact data by applying a negative power
     law (alpha) to its nonzero pixels, then interpolating on the zeroes using a
@@ -1164,197 +1308,6 @@ def distance_to_contact(D, alpha=1):
     return M
 
 
-def distance_law(matrix, indices=None, log_bins=True, base=1.1):
-    """Compute distance law as a function of the genomic coordinate aka P(s).
-    Bin length increases exponentially with distance if log_bins is True. Works
-    on dense and sparse matrices.
-
-    Parameters
-    ----------
-    matrix : numpy.array or scipy.sparse.coo_matrix
-        Hi-C contact map of the chromosome on which the distance law is
-        calculated.
-    indices : None or numpy array
-        List of indices on which to compute the distance law. For example
-        compartments or expressed genes.
-    log_bins : bool
-        Whether the distance law should be computed on exponentially larger
-        bins.
-
-    Returns
-    -------
-    numpy array of floats :
-        The start index of each bin.
-    numpy array of floats :
-        The distance law computed per bin on the diagonal
-    """
-
-    n = min(matrix.shape)
-    included_bins = np.zeros(n, dtype=bool)
-    if indices is None:
-        included_bins[:] = True
-    else:
-        included_bins[indices] = True
-    D = np.array(
-        [
-            np.average(matrix.diagonal(j)[included_bins[: n - j]])
-            for j in range(n)
-        ]
-    )
-    if not log_bins:
-        return np.array(range(len(D))), D
-    else:
-        n_bins = int(np.log(n) / np.log(base) + 1)
-        logbin = np.unique(
-            np.logspace(0, n_bins - 1, num=n_bins, base=base, dtype=np.int)
-        )
-        logbin = np.insert(logbin, 0, 0)
-        logbin[-1] = min(n, logbin[-1])
-        if n < logbin.shape[0]:
-            print("Not enough bins. Increase logarithm base.")
-            return np.array(range(len(D))), D
-        logD = np.array(
-            [
-                np.average(D[logbin[i - 1] : logbin[i]])
-                for i in range(1, len(logbin))
-            ]
-        )
-        return logbin[:-1], logD
-
-
-def distance_law_multi(
-    matrix,
-    frag_pos,
-    indices=None,
-    centro_pos=None,
-    log_bins=True,
-    base=1.1,
-    average=False,
-):
-    """Compute distance law as a function of the genomic coordinate aka P(s).
-    Bin length increases exponentially with distance if log_bins is True. Works
-    on dense and sparse matrices. Use this function if the genome is composed of
-    several chromosomes and provide a file with the positions of centromers
-
-    Parameters
-    ----------
-    matrix : numpy.array or scipy.sparse.csr_matrix
-        Hi-C contact map on which the distance law is calculated.
-    frag_pos : numpy.array
-        The list of fragments or bins start positions, in base pairs.
-    indices : numpy.array of int
-        Indices of bins that must be included. If None, all bins are used.
-        Useful to exclude trimmed bins or compartments.
-    centro_pos: numpy.array
-        List of centromers positions, in base pairs. The order must match those
-        of chromosomes in the fragments positions.
-    log_bins : bool
-        Whether the distance law should be computed on exponentially larger
-        bins. Default is True.
-    average : bool
-        If enabled, all P(s) computed will be averaged and a single array is
-        returned. By default, two list of arrays are returned; the first
-        contains the bins coordinates and the second the P(s) values. If only
-        frag pos is provided, each array will have C values where C is the number
-        of chromosomes. If centromeres is also given, each array will have 2*C
-        values. Chromosomes are returned in the same order as they appear in
-        frag_pos.
-
-    Returns
-    -------
-    list of numpy.array :
-        The start coordinate of each bin one array per chromosome or arm if not
-        averaged.
-    list of numpy.array :
-        The distance law computed per bin on the diagonal. One array per chrom-
-        osome or arm if not averaged
-    """
-
-    # Get bins where chromosomes start
-    chr_bins = np.where(frag_pos == 0)[0]
-    if centro_pos is not None:
-        # Sanity check: as many chroms as centromeres
-        if len(chr_bins) != len(centro_pos):
-            sys.stderr.write(
-                "ERROR: Number of chromosomes and centromeres differ."
-            )
-            sys.exit(1)
-        # Get bins of centromeres
-        centro_bins = np.zeros(len(centro_pos))
-        for i in range(len(chr_bins)):
-            if (i + 1) < len(chr_bins):
-                subfrags = frag_pos[chr_bins[i] : chr_bins[i + 1]]
-            else:
-                subfrags = frag_pos[chr_bins[i] :]
-            # index of last fragment starting before centro in same chrom
-            centro_bins[i] = chr_bins[i] + max(
-                np.where((subfrags // centro_pos[i]) == 0)[0]
-            )
-        # Combine centro and chrom bins into a single array. Values are start bins of arms
-        chr_bins = np.sort(np.concatenate((chr_bins, centro_bins)))
-
-    xs = [None] * len(chr_bins)
-    ps = [None] * len(chr_bins)
-    # Bins to use (e.g. to exclude trimmed bins or compartments)
-    if indices is None:
-        indices = np.array(range(matrix.shape[0]))
-    else:
-        indices = indices.astype(int)
-    # Iterate over chromosome / arm bin indices
-    for chrm in range(len(chr_bins)):
-        # Select bins from chrom / arm
-        if (chrm + 1) < len(chr_bins):
-            chr_start, chr_end = int(chr_bins[chrm]), int(chr_bins[chrm + 1])
-        else:
-            chr_start, chr_end = int(chr_bins[chrm]), matrix.shape[0]
-        # Subset indices and matrices. Shift indices to start in chrom
-        chr_idx_mask = np.where((indices >= chr_start) & (indices < chr_end))[
-            0
-        ]
-        usebins = np.array(indices[chr_idx_mask]) - chr_start
-        chr_mat = matrix[chr_start:chr_end, chr_start:chr_end]
-        # Get the ps for the different arms/chroms
-        chrom_xs, chrom_ps = distance_law(
-            chr_mat, indices=usebins, log_bins=log_bins, base=base
-        )
-        xs[chrm] = chrom_xs
-        ps[chrm] = chrom_ps
-
-    # Take the mean as P(s) and longest bin array
-    if average:
-        arr_len = np.array([len(x) for x in xs])
-        longest = np.where(arr_len == max(arr_len))[0][0]
-        xs = xs[longest]
-        ps = average_ps(ps)
-    return xs, ps
-
-
-def average_ps(ps):
-    """Computes the average of a several lists of floats of different lengths.
-    Values that do not exist in shorter list are not taken into account and do
-    not count as zeros.
-
-    Parameters
-    ----------
-    ps : list of lists of floats
-        The list of all the ps to average
-
-    Returns
-    -------
-    list of floats :
-        The averaged distance law
-    """
-    # Find longest chromosome / arm
-    max_length = max([len(i) for i in ps])
-    ps_values = np.zeros(max_length)
-    ps_occur = np.zeros(max_length)
-    for chrom_ps in ps:
-        ps_occur[: len(chrom_ps)] += 1
-        ps_values[: len(chrom_ps)] += chrom_ps
-    averaged_ps = ps_values / ps_occur
-    return averaged_ps
-
-
 def subsample_contacts(M, prop):
     """Bootstrap sampling of contacts in a sparse Hi-C map.
 
@@ -1934,7 +1887,7 @@ def compartments(M, normalize=True):
         N = np.copy(M)
     # Computation of genomic distance law matrice:
     dist_mat = np.zeros((n, n))
-    _, dist_vals = distance_law(N, log_bins=False)
+    _, dist_vals = distance_law_from_mat(N, log_bins=False)
     for i in range(n):
         for j in range(n):
             dist_mat[i, j] = dist_vals[abs(j - i)]
@@ -2013,7 +1966,7 @@ def compartments_sparse(M, normalize=True):
         N = copy.copy(M)
     N = N.tocoo()
     # Detrend by the distance law
-    dist_bins, dist_vals = distance_law(N, log_bins=False)
+    dist_bins, dist_vals = distance_law_from_mat(N, log_bins=False)
     N.data /= dist_vals[abs(N.row - N.col)]
     N = N.tocsr()
     # Make matrix symmetric (in case of upper triangle)
